
#include <pjsip/sip_generic_uri.h>
#include <pjsip/sip_msg.h>
#include <pjsip/sip_parser.h>
#include <pjsip/print_util.h>
#include <pj/pool.h>
#include <pj/assert.h>
#include <pj/string.h>
#include <pj/ctype.h>
#include <pj/except.h>
#include <pjlib-util/string.h>
#include <pjlib-util/scanner.h>

#ifdef __GNUC__
#  define HAPPY_FLAG	(void*)
#else
#  define HAPPY_FLAG
#endif

/* 
	A generic URI can consist of (For a complete BNF see RFC 2396):
	#?;:@&=+-_.!~*'()%$,/
*/
#define ALLOWED_CHARS "#?;:@&=+-_.!~*'()%$,/"

static pj_cis_buf_t cis_buf;
static pj_cis_t pjsip_GENERIC_URI_SPEC;

static const pj_str_t *generic_uri_get_scheme(const pjsip_generic_uri*);
static void *generic_uri_get_uri(pjsip_generic_uri*);
static pj_ssize_t generic_uri_print(
	pjsip_uri_context_e context,
	const pjsip_generic_uri *url, 
	char *buf, pj_size_t size);
static int generic_uri_cmp(
	pjsip_uri_context_e context,
	const pjsip_generic_uri *url1, 
	const pjsip_generic_uri *url2);
static pjsip_generic_uri* generic_uri_clone(
	pj_pool_t *pool, 
	const pjsip_generic_uri *rhs);



static void* generic_uri_parse(
	pj_scanner *scanner, 
	pj_pool_t *pool,
	pj_bool_t parse_params);

static pjsip_uri_vptr generic_uri_vptr = {
	HAPPY_FLAG &generic_uri_get_scheme,
	HAPPY_FLAG &generic_uri_get_uri,
	HAPPY_FLAG &generic_uri_print,
	HAPPY_FLAG &generic_uri_cmp,
	HAPPY_FLAG &generic_uri_clone
};


pj_status_t pjsip_generic_uri_subsys_init(void) {
	
	pj_status_t status;

	pj_cis_buf_init(&cis_buf);
	status = pj_cis_init(&cis_buf, &pjsip_GENERIC_URI_SPEC);
  PJ_ASSERT_RETURN(status==PJ_SUCCESS, status);
	pj_cis_add_alpha(&pjsip_GENERIC_URI_SPEC); /* a-zA-Z*/
 	pj_cis_add_num(&pjsip_GENERIC_URI_SPEC); /* 0-9 */
  pj_cis_add_str(&pjsip_GENERIC_URI_SPEC, ALLOWED_CHARS); /* #?;:@&=+-_.!~*'()%$,/ */

	status = pjsip_register_generic_uri_parser(&generic_uri_parse);
	PJ_ASSERT_RETURN(status==PJ_SUCCESS, status);
	
	return PJ_SUCCESS;
}


PJ_DEF(pjsip_generic_uri*) pjsip_generic_uri_create(pj_pool_t *pool) {
	pjsip_generic_uri *uri = (pjsip_generic_uri *)pj_pool_zalloc(pool, sizeof(pjsip_generic_uri));
  uri->vptr = &generic_uri_vptr;
  return uri;
}

static const pj_str_t *generic_uri_get_scheme( const pjsip_generic_uri *uri ) {
	return &(uri->scheme);
}

static void *generic_uri_get_uri( pjsip_generic_uri *uri ) {
    return uri;
}

static pj_ssize_t generic_uri_print(
	pjsip_uri_context_e context,
	const pjsip_generic_uri *uri, 
	char *buf, pj_size_t size) {

	char *startbuf = buf;
  char *endbuf = buf + size;

  PJ_UNUSED_ARG(context);
 
  /* Print scheme. */
  copy_advance(buf, uri->scheme);
  *buf++ = ':';

  /* Print uri. */
  copy_advance(buf, uri->uri);
  
	return (buf - startbuf);
}

static int generic_uri_cmp(
	pjsip_uri_context_e context,
	const pjsip_generic_uri *uri1, 
	const pjsip_generic_uri *uri2) {
	
	PJ_UNUSED_ARG(context);

	/* Scheme must match. */
	if(pj_strcmp(&(uri1->scheme), &(uri2->scheme)) != 0) {
		return -1;
	}

	/* URI must match. */
	if(pj_strcmp(&(uri1->uri), &(uri2->uri)) != 0) {
		return -1;
	}

	/* Equal. */
	return 0;
}

/* Clone *: URI */
static pjsip_generic_uri* generic_uri_clone(
	pj_pool_t *pool, 
	const pjsip_generic_uri *rhs) {
	
	pjsip_generic_uri *uri = pjsip_generic_uri_create(pool);
	pj_strdup(pool, &uri->scheme, &rhs->scheme);
	pj_strdup(pool, &uri->uri, &rhs->uri);

	return uri;
}

/* Parse *: URI 
 * THis actually returns (pjsip_generic_uri *) type. */

static void* generic_uri_parse(
	pj_scanner *scanner, 
	pj_pool_t *pool,
	pj_bool_t parse_params) {

	pjsip_generic_uri *uri = 0;
	const pjsip_parser_const_t *pc = pjsip_parser_const();
  int skip_ws = scanner->skip_ws;

	scanner->skip_ws = 0;
	
	uri = pjsip_generic_uri_create(pool); 

  pj_scan_get(scanner, &pc->pjsip_TOKEN_SPEC, &uri->scheme);
	if (pj_scan_get_char(scanner) != ':') {
		PJ_THROW(PJSIP_SYN_ERR_EXCEPTION);
	}
	
	pj_scan_get(scanner, &pjsip_GENERIC_URI_SPEC, &uri->uri);
	scanner->skip_ws = skip_ws;

	return uri;
}


